"""Interface with the Replit Database."""
import json
import os
from sys import stderr
from typing import Any, Callable, Dict, Tuple, Union

import requests


JSON_TYPE = Union[str, int, float, bool, type(None), dict, list]


class JSONKey:
    """Represents a key in the database that holds a JSON value.

    db.jsonkey() will initialize an instance for you,
    you don't have to do it manually.
    """

    __slots__ = ("db", "key", "dtype", "get_default", "discard_bad_data")

    def __init__(
        self,
        db: Any,
        key: str,
        dtype: JSON_TYPE,
        get_default: Callable = None,
        discard_bad_data: bool = False,
    ) -> None:
        """Initialize the key.

        Args:
            db (Any): An instance of ReplitDb
            key (str): The key to read
            dtype (JSON_TYPE): The datatype the key should be, can be typing.Any.
            get_default (Callable): A function that returns the default
                value if the key is not set. If it is None (the default) the dtype
                argument is used.
            discard_bad_data (bool): Don't prompt if bad data is read, overwrite it
                with the default. Defaults to False.
        """
        self.db = db
        self.key = key
        self.dtype = dtype
        self.get_default = get_default
        self.discard_bad_data = discard_bad_data

    def _default(self) -> JSON_TYPE:
        get_default_func = self.get_default or self.dtype
        return get_default_func()

    def _is_valid_type(self, data: JSON_TYPE) -> bool:
        return self.dtype is Any or isinstance(data, self.dtype)

    def _type_mismatch_msg(self, data: Any) -> str:
        return (
            f"Type mismatch: Got type {type(data).__name__},"
            "expected {self.dtype.__name__}"
        )

    def get(self) -> JSON_TYPE:
        """Get the value of the key.

        If an invalid JSON value is read or the type does not match, it will show a
            prompt asking the user what to do unless discard_bad_data is set.

        Returns:
            JSON_TYPE: The value read from the database
        """
        try:
            read = self.db[self.key]
        except KeyError:
            print(f"Database key {self.key} not set, setting it to default value")
            default = self._default()
            self.db[self.key] = default
            return default

        try:
            data = json.loads(read)
        except json.JSONDecodeError:
            return self._error("Invalid JSON data read", read)

        if not self._is_valid_type(data):
            return self._error(self._type_mismatch_msg(data), read,)
        return data

    def _error(self, error: str, read: str) -> JSON_TYPE:
        print(f"Error reading key {self.key!r}: {error}", file=stderr)
        if self.discard_bad_data:
            val = self._default()
            self.db[self.key] = json.dumps(val)
            print(f"Wrote default to key {self.key!r}")
            return val
        return self._should_discard_prompt(error, read)

    def _should_discard_prompt(self, error: str, read: str) -> bool:
        while True:
            choice = input(
                "d to use default, v to view the invalid data, c to insert custom "
                "value, ^C to exit: "
            )
            if choice.startswith("d"):
                print("Writing default...")
                val = self._default()
                self.db[self.key] = val
                return val
            elif choice.startswith("v"):
                print(f"Data read from key: {read!r}")
            elif choice.startswith("c"):
                toset = input(
                    f"Enter data to write, should be of type {self.dtype.__name__!r}"
                    " (leave blank to return to menu): "
                )
                if not toset:
                    continue
                try:
                    data = json.loads(toset)
                except json.JSONDecodeError:
                    print("Invalid JSON data!")
                else:
                    if not self._is_valid_type(data):
                        print(self._type_mismatch_msg(data))
                        continue

                    self.db[self.key] = toset
                    print("Wrote data to key")
                    return data

    def set(self, data: JSON_TYPE) -> None:
        """Set the value of the jsonkey.

        Args:
            data (JSON_TYPE): The value to set it to.

        Raises:
            TypeError: The type of the value set does not match the datatype.
        """
        if not self._is_valid_type(data):
            raise TypeError(self._type_mismatch_msg(data))
        self.db[self.key] = json.dumps(data)


class ReplitDb(dict):
    """Interface with the Replit Database."""

    __slots__ = ("db_url", "sess")

    def __init__(self, db_url: str) -> None:
        """Initialize database. You shouldn't have to do this manually.

        Args:
            db_url (str): Database url to use.
        """
        self.db_url = db_url
        self.sess = requests.Session()

    def __getitem__(self, key: str) -> str:
        """Get the value of an item from the database.

        Args:
            key (str): The key to retreive

        Raises:
            KeyError: Key is not set

        Returns:
            str: The value of the key
        """
        r = self.sess.get(f"{self.db_url}/{key}")
        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()
        return r.text

    def __setitem__(self, key: str, value: str) -> None:
        """Set a key in the database to value.

        Args:
            key (str): The key to set
            value (str): The value to set it to
        """
        r = self.sess.post(self.db_url, data={key: value})
        r.raise_for_status()

    def __delitem__(self, key: str) -> None:
        """Delete a key from the database.

        Args:
            key (str): The key to delete
        """
        r = self.sess.delete(f"{self.db_url}/{key}")
        r.raise_for_status()

    def keys(self, prefix: str = "") -> Tuple[str]:
        """Return all of the keys in the database.

        Args:
            prefix (str): The prefix the keys must start with,
                blank means anything. Defaults to "".

        Returns:
            Tuple[str]: The keys found.
        """
        r = requests.get(f"{self.db_url}", params={"prefix": prefix})
        r.raise_for_status()

        if not r.text:
            return tuple()
        else:
            return tuple(r.text.split("\n"))

    def to_dict(self, prefix: str = "") -> Dict[str, str]:
        """Dump all data in the database into a dictionary.

        Args:
            prefix (str): The prefix the keys must start with,
                blank means anything. Defaults to "".

        Returns:
            Dict[str, str]: All keys in the database.
        """
        keys = self.keys()
        data = {}
        for k in keys:
            data[k] = self[k]
        return data

    def values(self) -> Tuple[str]:
        """Get every value in the database.

        Returns:
            Tuple[str]: The values in the database.
        """
        data = self.to_dict()
        return tuple(data.values())

    def items(self) -> Tuple[Tuple[str]]:
        return self.to_dict().items()

    def jsonkey(
        self,
        key: str,
        dtype: JSON_TYPE,
        get_default: Callable = None,
        discard_bad_data: bool = False,
    ) -> JSONKey:
        """Initialize a JSONKey instance.

        A JSONKey is used to easily read and set JSON data from the database.
        Arguments are the same as JSONKey constructor.

        Args:
            key (str): The key to read
            dtype (JSON_TYPE): The datatype the key should be, can be typing.Any.
            get_default (Callable): A function that returns the default
                value if the key is not set. If it is None (the default) the dtype
                argument is used.
            discard_bad_data (bool): Don't prompt if bad data is read, overwrite it
                with the default. Defaults to False.

        Returns:
            JSONKey: The initialized JSONKey instance.
        """
        return JSONKey(
            db=self,
            key=key,
            dtype=dtype,
            get_default=get_default,
            discard_bad_data=discard_bad_data,
        )

    def __repr__(self) -> str:
        """A representation of the database.

        Returns:
            A string representation of the database object.
        """
        return f"<ReplitDb(db_url={self.db_url!r})>"


db_url = os.environ.get("REPLIT_DB_URL")
if db_url:
    db = ReplitDb(db_url)
else:
    print(
        "Warning: REPLIT_DB_URL does not exist, are we running on repl.it? "
        "Database will not function.",
        file=stderr,
    )
    db = None



"""The replit python module."""
from . import maqpy
from .audio import Audio
from .database import db


def clear() -> None:
    """Clear the terminal."""
    print("\033[H\033[2J", end="", flush=True)


audio = Audio()



"""Make apps quickly in python."""
import os

import flask
from werkzeug.local import LocalProxy

from . import files
from . import html
from .app import App
from .files import File
from .html import HTMLElement, Link, Page, Paragraph
from .utils import needs_params, needs_signin, sign_in_snippet, signin
from ..database import db

auth = LocalProxy(lambda: flask.request.auth)
signed_in = LocalProxy(lambda: flask.request.signed_in)

# TODO: signinwall(exclude=['/a', '/b'])
# TODO: @need_signin
# TODO: Param checking with @needs_params



import time
import unittest
import replit
from replit import audio, types
from replit.types import WaveType

test_file = '../test.mp3'


class TestAudio(unittest.TestCase):

    def test_creation(self):
        source = audio.play_file(test_file)
        self.assertEqual(source.path, test_file)
        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, 'Pausing Source')

    def test_pause(self):
        source = audio.play_file(test_file)
        source.volume = 2
        time.sleep(1)
        self.assertEqual(source.volume, 2, "Volume set to 2")

        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, 'Pausing Source')

        source.volume = .2
        time.sleep(1)
        self.assertEqual(source.volume, .2, 'Volume set to .2')

        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, 'Pausing Source')

    def test_loop_setting(self):
        source = audio.play_file(test_file)

        self.assertEqual(source.loops_remaining, 0, '0 loops remaining')
        source.set_loop(2)
        time.sleep(1)

        self.assertEqual(source.loops_remaining, 2, '2 loops remaining')
        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, 'Pausing Source')

    def test_other(self):
        source = audio.play_file(test_file)

        self.assertIsNotNone(source.end_time)
        self.assertIsNotNone(source.start_time)
        self.assertIsNotNone(source.remaining)
        source.paused = True
        time.sleep(1)
        self.assertEqual(source.paused, True, 'Pausing Source')

    def test_tones(self):
        try:
            audio.play_tone(2, 400, 2)
        except TimeoutError or ValueError as e:
            self.fail(e)


if __name__ == '__main__':
    unittest.main()



"""Utitilities to make development easier."""
from functools import wraps
from typing import Any, Callable, Union

import flask

from .html import Page


sign_in_snippet = (
    '<script authed="location.reload()" '
    'src="https://auth.turbio.repl.co/script.js"></script>'
)


def signin(title: str = "Please Sign In") -> Page:
    """Return a sign-in page.

    Args:
        title (str): The title of the sign in page. Defaults to "Please Sign In".

    Returns:
        Page: The sign-in page.
    """
    return Page(title=title, body=sign_in_snippet)


def needs_signin(func: Callable = None, login_html: str = sign_in_snippet) -> Callable:
    """A decorator that enforces that the user is signed in before accessing the page.

    Args:
        func (Callable): The function passed in if used as a decorator. Defaults to
            None.
        login_html (str): The HTML to show when the user needs to sign in. Defaults to
            sign_in_snippet.

    Returns:
        Callable: The new handler.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def handler(*args: Any, **kwargs: Any) -> flask.Response:
            if flask.request.signed_in:
                return func(*args, **kwargs)
            else:
                return login_html

        return handler

    if func is not None:  # called with no options @needs_signin
        return decorator(func)
    else:  # called with options, eg @needs_signin(login_html='...')
        return decorator


def needs_params(
    *param_names: str,
    src: Union[str, dict] = "form",
    onerror: Callable[[str], flask.Response] = None,
) -> Callable:
    """Require paramaters before a handler can be activated.

    Args:
        param_names (str): The paramaters that must be in the request.
        src (Union[str, dict]): The source to get the paramaters from. Can be "form"
            to use flask.request.form (POST requests), "query" for flask.request.query
            (GET requests), or a custom dictionary.
        onerror (Callable): A function to handle when a paramater is missing. It will
            be passed the parameter that is missing. If no function is specified a
            handler that returns a descriptive error and 400 Bad Request status code
            will be used.

    Raises:
        TypeError: No paramaters were provided or an invalid one was provided.

    Returns:
        Callable: The new handler.
    """
    if len(param_names) < 1:
        raise TypeError("You must specify at least one required paramater name")
    # If function is used as a decorator with no arguments, the first argument will be
    # a function, so type check all of the param names to catch mistakes
    if not all(isinstance(p, str) for p in param_names):
        raise TypeError("All paramater names should be strings.")

    def default_onerror(missing_param: str) -> flask.Response:
        return flask.make_response(
            f"Parameter {missing_param!r} is required but is missing",
            400,
            mimetype="text/plain",
        )

    onerror = default_onerror if onerror is None else onerror

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def handler(*args: Any, **ignoredkwargs: Any) -> flask.Response:
            if src == "form":
                params = flask.request.form
            elif src == "query":
                params = flask.request.args
            else:
                params = src
            param_kwargs = {}
            for p in param_names:
                if p not in params:
                    return onerror(p)
                param_kwargs[p] = params[p]
            return func(*args, **param_kwargs)

        return handler

    return decorator



from typing import List
from typing_extensions import TypedDict
from enum import Enum


class ReaderType(Enum):
    'An Enum for the types of sources.'

    def __str__(self) -> str:
        return self._value_

    def __repr__(self) -> str:
        return f'ReaderType.{self._name_}'

    wav_file = 'wav'
    'ReaderType : The type for a .wav file.'
    aiff_file = 'aiff'
    'ReaderType : The type for a .aiff file.'
    mp3_file = 'mp3'
    'ReaderType : The type for a .mp3 file.'
    tone = 'tone'
    'ReaderType : The type for a generated tone.'


class WaveType(Enum):
    'The different wave shapes that can be used for tone generation.'

    def __str__(self) -> str:
        return self._value_

    WaveSine = 0
    'WaveType : The WaveSine wave shape.'
    WaveTriangle = 1
    'WaveType : The Triangle wave shape.'
    WaveSaw = 2
    'WaveType : The Saw wave shape.'
    WaveSqr = 3
    'WaveType : The Square wave shape.'


file_types: List[ReaderType] = [ReaderType.aiff_file,
                                ReaderType.wav_file, ReaderType.mp3_file]
'The different file types for sources in a list.'


class RequestArgs(TypedDict, total=False):
    'The additional arguments for a request that are type-specific.'
    Pitch: float
    'float : The pitch/frequency of the tone. Only used if the request type is tone.'
    Seconds: float

    'float : The duration for the tone to be played. Only used if the request type is tone.'
    WaveType: WaveType or int
    'WaveType : The wave type of the tone. Only used if the request type is tone.'
    Path: str
    'str : The path to the file to be read. Only used if the request is for a file type.'


class RequestData(TypedDict):
    'A request to pid1 for a source to be played.'
    ID: int
    'int : The ID of the source. Only used for updating a pre-existing source.'
    Paused: bool or None
    'bool or None : Wether the source with the provided ID should be paused or not.  Can only be used when updating a source.'
    Volume: float
    'float : The volume the source should be played at. (1 being 100%)'
    DoesLoop: bool
    'bool : Wether the source should loop / repeat or not. Defaults to false.'
    LoopCount: int
    'int : How many times the source should loop / repeat.  Defaults to 0.'
    Name: str
    'str : The name of the source.'
    Type: ReaderType or str
    'ReaderType : The type of the source.'
    Args: RequestArgs
    'RequestArgs : The additional arguments for the source.'


class SourceData(TypedDict):
    '''A source's raw data, as a payload.'''
    Name: str
    'str : The name of the source.'
    Type: str
    'str : The type of the source.'
    Volume: float
    'float : The volume of the source.'
    Duration: int
    'int : The duration of the source in milliseconds.'
    Remaining: int
    'int : How many more milliseconds the source will be playing.'
    Paused: bool
    'bool : Wether the source is paused or not.'
    Loop: int
    'int : How many times the source will loop. If 0, the source will not repeat itself.'
    ID: int
    'int : The ID of the source.'
    EndTime: str
    'str : The estimated timestamp for when the source will finish playing.'
    StartTime: str
    'str : When the source started playing.'
    Request: RequestData
    'RequestData : The request used to create the source.'


class AudioStatus(TypedDict):
    'The raw data read from /tmp/audioStatus.json.'
    Sources: List[SourceData] or None
    'List[SourceData] : The sources that are know to the audio manager.'
    Running: bool
    'bool : Wether the audio manager knows any sources or not.'
    Disabled: bool
    'bool : Wether the audio manager is disabled or not.'



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['replit', 'replit.audio', 'replit.database', 'replit.maqpy']

package_data = \
{'': ['*']}

install_requires = \
['flask>=1.1.2,<2.0.0',
 'requests>=2.24.0,<3.0.0',
 'typing_extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'replit',
    'version': '1.4.0',
    'description': 'A library for interacting with features of repl.it',
    'long_description': '# replit-py\n\nReplit-py is a python library designed to be run from a repl on [repl.it](https://repl.it).\n\n### Features\n\n- Fully featured database client for Repl DB\n- Audio library which can play tones and files\n- Clear the terminal\n\n### Documentation\n\nThe documentation can be found [here](https://replit-python-docs.scoder12.repl.co).\n',
    'author': 'mat',
    'author_email': 'pypi@matdoes.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/replit/replit-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
